//---------------------------------------------------------------- -o--
// GTHelper.java 
//  
// Class methods and constants shared between GalacticTransaction, 
//   GTTest* PuzzleTest classes.
// 
//
// TEST CONVENTIONS  
//   . All input files live in TESTFILES_DIR
//   . All output files written to TESTFILESPREV_DIR
//   . TESTFILESPREV_DIR is deleted before tests are rerun
//   . Input files are named as follows:
//       "section[123]-<expectation>-<thingTested>-input.txt"
//   . Output files have the same name, but with trailing "-output.txt" 
//
// TEST PROCESS
//   . GalacticTransaction is run over streams derived from input and
//       output files
//   . Output file data is suffixed with "\nRVAL = ??" where "??" is
//       replaced with the return value of GalacticTransaction
//   . If output file does not exist, the test fails, otherwise the
//       output is converted into a string
//   . The string is matched against expected test results derived
//       from knowledge of the input file and relvant parse code
//
//
// METHODS--
//   setFilesAndRunProgram
//   stringContains
//   dumpResultsString
//   tokenCount
//   floatAsString
//

import java.lang.*;
import java.io.*;
import java.util.*;

import junit.framework.*;
import org.junit.*;
import org.apache.commons.io.*;
import org.hamcrest.*;



//----------------------------------------------------------------- -o--
public class GTHelper
{
  public static final String  
      TESTFILES_DIR      = "testfiles/",
      TESTFILESPREV_DIR  = TESTFILES_DIR + "/previousRun/";

  static final String  BADINPUT_PREFIX   = "ILLEGAL INPUT LINE: ";
  static final String  INCOMPLETE_PARSE  = "INCOMPLETE PARSE: ";
  static final String  PARSE_ERROR       = "PARSE ERROR: ";
  static final String  INPUT_ERROR       = "INPUT ERROR: ";


  private static final String  _CLASSNAME = "GTHelper";



  //-------------------------------------------------------------- -o--
  // Methods.
  //

  //------------------------------ -o-
  // setFilesAndRunProgram
  //
  // Use name of input file to create input and output streams.
  // Run GalacticTransaction to generate an output file for testing.
  //
  // Input files are crafted to provoke output errors from the parse 
  //   logic of Section1 of a Galactic Transaction.
  // 
  //
  // NB  Error checking is minimal compared to the "application"
  //     interface of Puzzle.  Gross problems caught via
  //     try-catch -- otherwise ASSUMED that test developer knows what
  //     they are doing.
  //
  public static String  setFilesAndRunProgram(String inputFileBasename)
  {
    String  outputFileBasename = inputFileBasename.replace("input", "output");

    String  inputFileFullPathname   = TESTFILES_DIR     + inputFileBasename;
    String  outputFileFullPathname  = TESTFILESPREV_DIR + outputFileBasename;

    BufferedReader inputStream;
    BufferedWriter outputStream;


    //
    try {
      inputStream = new BufferedReader(
                      new FileReader(new File(inputFileFullPathname)));
      outputStream = new BufferedWriter(
                      new FileWriter(new File(outputFileFullPathname)));


      GalacticTransaction galacticTransaction = 
          new GalacticTransaction(inputStream, outputStream, false);
      Integer rval = galacticTransaction.main();

      String  rvalString  = "\nRVAL = "+ rval +"\n";
      outputStream.write(rvalString, 0, rvalString.length());

      
      inputStream.close();

      outputStream.flush();
      outputStream.close();


    } catch (Exception e) {
      System.err.println("TEST ERROR in "+ _CLASSNAME +": "+ e.toString());
    }


    return outputFileFullPathname;

  } // setFilesAndRunProgram()


  //------------------------------ -o-
  // stringContains
  // 
  // RETURN  true if source CONTAINS pattern; 
  //         false, otherwise.
  //
  public static Boolean  stringContains(String source, String pattern) 
  {
    Scanner  scanner   = new Scanner(source);
    String   matchPtr  = scanner.findInLine(pattern);

    return (null != matchPtr);
  } 


  //------------------------------ -o-
  public static void  dumpResultsString( String outputResults, 
                                         String inputFileName )
  {
    System.out.println(
        "\n:: RESULTS OF PARSING INPUT FILE "+ inputFileName +"--" );
    System.out.println(outputResults);
    System.out.println();
  }


  //------------------------------ -o-
  public static Integer  tokenCount(String line)
  {
    StringTokenizer  st = new StringTokenizer(line);
    return st.countTokens();
  }


  //------------------------------ -o-
  // floatAsString
  //
  // Convert Float to a String, removing decimal fraction if it is zero. 
  //
  public static String  floatAsString(Float floatValue)
  {
    Float  integerValue = new Float( floatValue.intValue() );

    if (floatValue.equals(integerValue)) {
      return ("" + floatValue.intValue());
    } else {
      return ("" + floatValue);
    }
  }

} // class GTHelper

