//---------------------------------------------------------------- -o--
// GalacticTransaction.java
// 
// GalacticTransaction solves a puzzle titled "MERCHANT'S GUIDE TO THE GALAXY".
//
//
// PROBLEM SUMMARY--
//   A Galactic Transaction statement defines values for foreign currency 
//   and materials for sale, then asks for a translation of foreign currency
//   values and the cost of materials expressed in the foreign currency.
//   
//   The foreign currency is expressed in Galactic Numerals, the rules for 
//   which are governed by Roman Numerals.  The value of Material Units is
//   expressed in ratios between Galactic Numerals and Arabic numerals.  
//   Knowing these values enables one to answer questions about the cost in
//   'Credits' for materials expressed in Galactic Numerals.
//   
//   A Galactic Transaction has a fixed grammar of optional and expandable
//   sections.  The first two sections define new terms and values for 
//   Galactic Numbers and Material Units -- the names of Galactic Numbers 
//   and Material Units are not known in advance and must be determined when
//   parsing the Galactic Transaction.  The third and final section asks
//   questions about the Arabic numeral value of Galactic Numerals and the 
//   cost in Credits of materials expressed in Galactic Numerals.
//   
//
// METHODS--
//   init
//   GalacticTransaction
//   outputError
//
//   parseSectionOne
//   parseSectionTwo
//   parseSectionThree
//   main
//
//
// @version 0.1 
//

import java.lang.*;
import java.io.*;
import java.util.*;



public class  GalacticTransaction 
{                                                       
  // Arguments input to the parse.
  //
  private BufferedReader  inputStream;
  private BufferedWriter  outputStream;


  // When false, stdoutIsAllowed switch redirects error messages
  //   to outputStream for use in testing.  
  //
  private Boolean  stdoutIsAllowed = true;


  // Parse state indicates the current section of Galactic Transaction.
  //
  private enum GTSection  { ONE, TWO, THREE };
  private GTSection section = GTSection.ONE;

  private Boolean  atLeastOneLineSuccessfullyParsedInSection1 = false;


  // Fixed tokens in the namespace of a Galactic Transaction.
  //
  private static final String  TOKEN_IS        = "is";
  private static final String  TOKEN_CREDITS   = "Credits";
  private static final String  TOKEN_HOWMUCH   = "how much is ";
  private static final String  TOKEN_HOWMANY   = "how many Credits is ";
  private static final String  TOKEN_QUESTION  = "?";

  private static final String  CANONICAL_ERROR = 
				 "I have no idea what you are talking about";

  //
  private static Integer  errorCount = 0;

  private static Integer  verbose = 0;



  //-------------------------------------------------------------- -o--
  // Contructors.
  //

  //---------------------- -o-
  private void init( BufferedReader bufferedReader, 
                     BufferedWriter bufferedWriter )  
  {
    inputStream   = bufferedReader;
    outputStream  = bufferedWriter;
  }


  //---------------------- -o-
  GalacticTransaction( BufferedReader bufferedReader, 
                       BufferedWriter bufferedWriter )  
  { 
    init(bufferedReader, bufferedWriter);
  } 


  GalacticTransaction( BufferedReader bufferedReader, 
                       BufferedWriter bufferedWriter,
                       Boolean _stdoutIsAllowed )
  { 
    stdoutIsAllowed = _stdoutIsAllowed;

    init(bufferedReader, bufferedWriter);
  } 
 


  //-------------------------------------------------------------- -o--
  // Helper methods.
  //

  //--------------- -o-
  // outputError
  //
  // Output is sqeulched when running test cases.  
  // Capture errors in the output file for testing.
  //
  void outputError(String s)  throws Exception
  {
    errorCount += 1;

    if (stdoutIsAllowed) {
      System.err.println(s);
    } else {
      outputStream.write(s, 0, s.length());
      outputStream.newLine();
    }
  }



  //-------------------------------------------------------------- -o--
  // Methods for parsing Galactic Transactions.
  //

  //--------------- -o-
  String parseSectionOne(String line)  throws Exception
  {
    String[]  outerTokens;
    Integer   tokenCount;


    // Sanity check input.
    // Split the line to isolate GalacticNumber and RomanNumber.
    //
    tokenCount = GTHelper.tokenCount(line);

    if (3 != tokenCount)
    {
      if ((tokenCount > 3) && atLeastOneLineSuccessfullyParsedInSection1)
      {
        section = GTSection.TWO;
        return line;
      } 

      outputError(GTHelper.BADINPUT_PREFIX + line +" -- Too few tokens.");
      return null;
    }


    outerTokens = line.split("\\s"+ TOKEN_IS +"\\s");

    if ( (2 != outerTokens.length)
           || (outerTokens[0].length() <= 0) 
              || (outerTokens[1].length() <= 0) )
    {
      outputError(GTHelper.BADINPUT_PREFIX + line +" -- Missing tokens.");
      return null;
    }


    //
    try {
      GalacticNumber.mapFromRomanNumber(
        			outerTokens[0].trim(), outerTokens[1].trim());

    } catch (GTException e) {
      outputError(GTHelper.BADINPUT_PREFIX + line 
          + " -- Bad mapping of GalacticNumber to RomanNumber.");
      return null;
    }

    atLeastOneLineSuccessfullyParsedInSection1 = true;


    return null;

  } // parseSectionOne()


  //--------------- -o-
  String parseSectionTwo(String line)  throws Exception
  {
    String[]  outerPhrases,
              firstHalfTokens,
	      secondHalfTokens;

    String  materialUnitName,
            galacticNumeralStr;

    Float   arabicNumeral;


    // Sanity check input:
    //   . Defer to Section 3
    //   . Split the line
    //
    if (line.startsWith(TOKEN_HOWMUCH) || line.startsWith(TOKEN_HOWMANY))
    {
      section = GTSection.THREE;
      return line;
    }

    //
    outerPhrases = line.split("\\s"+ TOKEN_IS +"\\s");

    if ( (2 != outerPhrases.length)
           || (outerPhrases[0].length() <= 0) 
              || (outerPhrases[1].length() <= 0) )
    {
      outputError(GTHelper.BADINPUT_PREFIX + line +" -- Missing tokens.");
      return null;
    }


    // Parse first 1/2 of line.
    //
    if (2 > GTHelper.tokenCount(outerPhrases[0]))
    {
      outputError(
        GTHelper.BADINPUT_PREFIX + line + " -- Missing tokens in first half.");
      return null;
    }

    firstHalfTokens = outerPhrases[0].split("\\s");

    //
    materialUnitName = firstHalfTokens[firstHalfTokens.length - 1].trim();

    galacticNumeralStr = "";
    for (int i = 0; i < (firstHalfTokens.length - 1); i += 1)
    {
      galacticNumeralStr += " " + firstHalfTokens[i].trim();
    }

    galacticNumeralStr = galacticNumeralStr.trim();


    // Parse second 1/2 of line.
    //
    if (2 != GTHelper.tokenCount(outerPhrases[1]))
    {
      outputError(
        GTHelper.BADINPUT_PREFIX + line 
	    + " -- Second half must contain two tokens.");
      return null;
    }

    secondHalfTokens = outerPhrases[1].split("\\s");

    if (! secondHalfTokens[1].equals(TOKEN_CREDITS)) 
    {
      outputError(
        GTHelper.BADINPUT_PREFIX + line 
	    + " -- Second half last token must be \""+ TOKEN_CREDITS +"\".");
      return null;
    }

    try {
      arabicNumeral = new Float(secondHalfTokens[0].trim());

    } catch (NumberFormatException e) {
      outputError(
        GTHelper.BADINPUT_PREFIX + line 
	  + " -- Bad Arabic Numeral in second half.  ("
	    + secondHalfTokens[0] + ")" );
      return null;
    }


    // Create MaterialUnit.
    //
    try {
      MaterialUnit.mapFromGalacticAndArabicNumerals(
      			materialUnitName, galacticNumeralStr, arabicNumeral);

    } catch (GTException e) {
      outputError(GTHelper.BADINPUT_PREFIX + line +" -- "+ e);
      return null;
    }


    return null;
  } // parseSectionTwo()


  //--------------- -o-
  String parseSectionThree(String line)  throws Exception
  {
    String    remainingString,
	      arabicNumeralStr,
              outputAnswer,
	      materialUnitStr;

    String[]  tokens;

    Float     creditValueOfMaterialUnit;

    GalacticNumeral  gn = null;


    // 
    if (line.startsWith(TOKEN_HOWMUCH))
    {
      remainingString = line.replace(TOKEN_HOWMUCH, "");

      if (2 > GTHelper.tokenCount(remainingString)) {
	outputError(
	  GTHelper.BADINPUT_PREFIX + line 
	    + " -- HOW MUCH remainder must have at least two tokens.");
	return null;
      }

      tokens = remainingString.split("\\s");

      if (! tokens[tokens.length-1].trim().equals(TOKEN_QUESTION))
      {
	outputError(
	  GTHelper.BADINPUT_PREFIX + line 
	    +" -- HOW MUCH last token must be \""+ TOKEN_QUESTION +"\".");
	return null;
      }

      remainingString = remainingString.replace(TOKEN_QUESTION, "");


      // Parse Galactic Numeral.
      // Remove radix from floats lacking fractional portion.
      // Generate output with answer to HOWMUCH question.
      //
      try {
	gn = new GalacticNumeral(remainingString);

      } catch (GTException e) {
	outputError( 
	  GTHelper.BADINPUT_PREFIX + line + " -- HOW MUCH: " + e); 
	return null;

      }

      arabicNumeralStr = GTHelper.floatAsString(gn.asArabicNumeral());

      outputAnswer = 
          remainingString.trim() +" "+ TOKEN_IS +" "+ arabicNumeralStr;


    } else if (line.startsWith(TOKEN_HOWMANY)) 
    {
      remainingString = line.replace(TOKEN_HOWMANY, "");

      if (3 > GTHelper.tokenCount(remainingString)) {
	outputError(
	  GTHelper.BADINPUT_PREFIX + line 
	    + " -- HOW MANY remainder must have at least three tokens.");
	return null;
      }

      tokens = remainingString.split("\\s");

      if (! tokens[tokens.length-1].trim().equals(TOKEN_QUESTION))
      {
	outputError(
	  GTHelper.BADINPUT_PREFIX + line 
	    +" -- HOW MANY last token must be \""+ TOKEN_QUESTION +"\".");
	return null;
      }

      remainingString = remainingString.replace(TOKEN_QUESTION, "");


      // Parse "<GalacticNumeral> <MaterialUnit>"
      // Remove radix from floats lacking fractional portion.
      // Generate output with answer to HOWMANY question.
      //
      materialUnitStr = tokens[tokens.length-2].trim();

      creditValueOfMaterialUnit = 
        MaterialUnit.asArabicNumeral(materialUnitStr);

      if (null == creditValueOfMaterialUnit)
      {
	outputError( 
	  GTHelper.BADINPUT_PREFIX + line 
	    + " -- HOW MANY: No such MaterialUnit.  (" + materialUnitStr +")");
	return null;
      }

      remainingString = remainingString.replace(materialUnitStr, "");


      try {
	gn = new GalacticNumeral(remainingString);

      } catch (GTException e) {
	outputError( 
	  GTHelper.BADINPUT_PREFIX + line + " -- HOW MANY: " + e); 
	return null;

      }


      arabicNumeralStr = GTHelper.floatAsString(
        gn.asArabicNumeral() * MaterialUnit.asArabicNumeral(materialUnitStr));

      outputAnswer = 
	  remainingString.trim() +" "+ materialUnitStr 
	     +" "+ TOKEN_IS +" "+ arabicNumeralStr +" "+ TOKEN_CREDITS;


    } else {
      outputAnswer = CANONICAL_ERROR;
    }


    //
    outputStream.write(outputAnswer, 0, outputAnswer.length());
    outputStream.newLine();
      

    return null;

  } // parseSectionThree()




  //-------------------------------------------------------------- -o--
  Integer main()  throws Exception
  {
    String   line = null, 
             reparseLine = null;

    errorCount = 0;   // 0 is success; other values indicate failure



    // Read each line of the input file, passing each to methods that
    //   parse the three sections of a Galactic Transaction.
    //
    // If a given section cannot parse a line, it may return it for the
    //   next section.  
    //
    try {
      for ( line = inputStream.readLine();
            null != line;
            line = (null != reparseLine) ? reparseLine : inputStream.readLine())
      {
        if (verbose > 0) 
	{
	  if (stdoutIsAllowed) {
	    if (null != reparseLine) {
	      System.out.println("REPARSE: " + line);
	    } else {
	      System.out.println("LINE: " + line);
	    }
	  }
	}

        switch (section) {
        case ONE:
          reparseLine = parseSectionOne(line);
          break;

        case TWO:
          reparseLine = parseSectionTwo(line);
          break;

        case THREE:
          reparseLine = parseSectionThree(line);
          break;
        }
      }

    } catch (GTException e) {
      outputError(GTHelper.PARSE_ERROR + e.getMessage());
    }


    // Post-condition error cases.
    //
    if (null != line) {
      outputError(GTHelper.INCOMPLETE_PARSE +"Some input lines remain unread.");
    }


    outputStream.flush();


    return errorCount;
  } // main()

}  // public class GalacticTransaction 

