//---------------------------------------------------------------- -o--
// RomanNumber.java
// 
// Emeration class to hold the value of Roman Number and characteristics
//   necessary to parse Roman Numeral Expressions. 
//
// All methods other than asArabicNumeral() created to enable
//   RomanNumeral to parse sequences of Roman Numbers.
//
// Per Rule #3 of "Rules for the Ordering of Roman Numbers", two roles
// of a RomanNumber are mutually exclusive:
//   1) acting as a subtractor (hasSubtracted) 
//   2) performing addition against self (numberOfConsecutiveOccurences) 
//
//
// CLASS METHODS--
//   isNumber
//
// INSTANCE METHODS--
//   init
//   RomanNumber
//   asArabicNumeral
//   getIsAvailable
//   setIsAvailable
//
//   maySubtractFrom
//   setSubtracted
//   addOccurence
//   resetToDefaults
//
//   isLessThan
//   isGreaterThan
//
//
// @version 0.1 
//

import java.util.*;



//-------------------------------------------------------------- -o--
public enum RomanNumber 
{                                               
  M  ( 1000                               ),
  D  (  500                               ),
  C  (  100, RomanNumber.M, RomanNumber.D ),
  L  (   50                               ),
  X  (   10, RomanNumber.C, RomanNumber.L ),
  V  (    5                               ),
  I  (    1, RomanNumber.X, RomanNumber.V ) ;



  //-------------------------------------------------------------- -o--
  // Properties.
  //

  private Boolean  
    isAvailable    = true,   // is available to RomanNumeral parse
    hasSubtracted  = false;  // has subtracted from another RomanNumber

  protected HashSet<RomanNumber>  subtracteeSet = null;
      // Set of RomanNumbers that may be subtracted-from by this one

  private Integer  numberOfConsecutiveOccurences = 0;


  //
  private Float  arabicNumeralValue = new Float(0);



  //-------------------------------------------------------------- -o--
  // Constructors.
  //
  
  //--------------- -o-
  private void init(Integer _arabicNumeralValue)
  {
    arabicNumeralValue = new Float(_arabicNumeralValue);
  }


  //--------------- -o-
  RomanNumber(Integer _arabicNumeralValue)
  { 
    init(_arabicNumeralValue);
  }
  
  
  //--------------- -o-
  RomanNumber(Integer _arabicNumeralValue, RomanNumber rn1, RomanNumber rn2)
  {
    init(_arabicNumeralValue);

    subtracteeSet = new HashSet<RomanNumber>();
    subtracteeSet.add(rn1);
    subtracteeSet.add(rn2);
  }



  //-------------------------------------------------------------- -o--
  // Class methods.
  //

  //--------------- -o-
  static Boolean  isNumber(String token)
  {
    try {
      RomanNumber rn = RomanNumber.valueOf(token);
    } catch (IllegalArgumentException  e) {
      return false;
    }

    return true;
  }



  //-------------------------------------------------------------- -o--
  // Instance methods.
  //

  //--------------- -o-
  Float  asArabicNumeral()  { return arabicNumeralValue; }


  //--------------- -o-
  Boolean  getIsAvailable()                   { return isAvailable; }
  void     setIsAvailable(Boolean boolValue)  { isAvailable = boolValue; }


  //--------------- -o-
  Boolean  maySubtractFrom(RomanNumber followingRN)
  {
    if ( (null == subtracteeSet) 
           || hasSubtracted || (numberOfConsecutiveOccurences > 0) )
    {
      return false;
    }

    if (null == followingRN) { 
      return false;
    }

    for (RomanNumber rn : subtracteeSet) {
      if (followingRN.equals(rn)) {
        return true;
      }
    }

    return false;
  }


  //--------------- -o-
  void  setSubtracted() 
  {
    hasSubtracted  = true;
    isAvailable    = false;             // Ordering Roman Numbers: Rule #3
  }


  //--------------- -o-
  void  addOccurence() 
  {
    numberOfConsecutiveOccurences += 1;

    if (numberOfConsecutiveOccurences >= 3) {
      isAvailable = false;              // Ordering Roman Numbers: Rule #1
    }
  }


  //--------------- -o-
  void  resetToDefaults()
  {
    isAvailable    = true;   
    hasSubtracted  = false; 

    numberOfConsecutiveOccurences = 0;
  }


  //--------------- -o-
  Boolean  isLessThan(RomanNumber rn)
  {
    return (arabicNumeralValue < rn.asArabicNumeral());
  }


  //--------------- -o-
  Boolean  isGreaterThan(RomanNumber rn)
  {
    return (arabicNumeralValue > rn.asArabicNumeral());
  }

}  // public class RomanNumber 

